/****************************************************************************
* abstract_config.h
*
* Platform independent wrapper for config files
*
* Copyright (C) 2004 Rafal Mantiuk <rafm@users.sourceforge.org>
* 'describeKey' function is derived from the OpieReader sources.
*
* This program is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*
****************************************************************************/

#ifndef ABSTRACT_CONFIG_H
#define ABSTRACT_CONFIG_H

#include <qstring.h>
#include <qstringlist.h>

#ifdef DESKTOP
class Registry;
#else
class Config;
#endif


class AbstractConfig
{
public:
  AbstractConfig( const char *name );
  
  ~AbstractConfig();
  
  void writeEntry ( const char *key, const char * value );  
  void writeEntry ( const char *key, int num );
  void writeEntry ( const char *key, const QString value );
  void writeEntry ( const char *key, const QStringList &stringList, const char sep );
  void writeEntry ( const char *key, bool value );

  QString readEntry( const char *key, const char *deflt = NULL ) const;
  int readNumEntry( const char *key, int deflt = -1 ) const;
  bool readBoolEntry( const char *key, bool deflt = false ) const;
  QStringList readListEntry( const char *key, const char sep ) const;
  
  void setGroup( const char *gname );
private:
#ifdef DESKTOP
  Registry *registry;
#else
  Config *config;
#endif
  
};

QStringList escapeSafeSplit( const QChar & sep, const QString & str );
QString escapeString( QString str, QString characters );
QString unescapeString( QString str, QString characters );

#endif
