#include <qlayout.h>
#include <qlineedit.h>
#include <qmultilineedit.h>
#include <qpushbutton.h>
#include <qhbox.h>
#include <qvbox.h>
#include <qlabel.h>

#ifndef DESKTOP
#include "zfiledialog.h"
#else
#include <qfiledialog.h>
#endif

#include "dicproperties.h"

DicProperties::DicProperties( QWidget *parent, DicPropertiesMode mode, const QString &dicPath ) :
  QDialog( parent, "dic_properties", true ),
  fileNameEdited( false ),
  mode( mode ),
  dicPath( dicPath )
{
  setCaption( tr( "Create new dictionary" ) );

  QVBox *vbox;
  QHBox *hbox;
  
  QBoxLayout *l = new QVBoxLayout( this );
  l->setSpacing( 3 );
  l->setMargin( 4 );

  hbox = new QHBox(this);
  l->addWidget( hbox );
  hbox->setSpacing(4);

  vbox = new QVBox(hbox);
  vbox->setSpacing(4);

  QLabel *label = new QLabel(tr("Name"), vbox);
  label = new QLabel(tr("File"), vbox);

  vbox = new QVBox(hbox);
  vbox->setSpacing(4);
  
  nameLE = new QLineEdit( vbox );
  hbox = new QHBox(vbox);
  hbox->setSpacing(2);
  fileLE = new QLineEdit( hbox );
  QPushButton *fileSelBt = new QPushButton( "..", hbox );
  connect( fileSelBt, SIGNAL(clicked()), 
    this, SLOT(onFileSelClicked()));
  
  vbox = new QVBox(this);
  l->addWidget( vbox );
  hbox->setSpacing(2);
  label = new QLabel(tr("Description"), vbox);
  descriptionMLE = new QMultiLineEdit( vbox );

  hbox = new QHBox(this);
  l->addWidget( hbox );
  hbox->setSpacing(4);

  vbox = new QVBox(hbox);
  vbox->setSpacing(4);
  label = new QLabel(tr("Maintainer"), vbox);
  label = new QLabel(tr("Copyright"), vbox);

  vbox = new QVBox(hbox);
  vbox->setSpacing(4);
  
  maintainerLE = new QLineEdit( vbox );
  copyrightLE = new QLineEdit( vbox );

#ifdef DESKTOP
  hbox = new QHBox( this );
  l->addWidget( hbox );
  hbox->layout()->addItem( new QSpacerItem( 0, 0, QSizePolicy::Minimum, QSizePolicy::Minimum ) );
  QButton *doneBt = new QPushButton( tr( "Done" ), hbox );
  doneBt->setSizePolicy( QSizePolicy( QSizePolicy::Maximum, QSizePolicy::Maximum ) );
  QButton *cancelBt = new QPushButton( tr( "Cancel" ), hbox );  
  cancelBt->setSizePolicy( QSizePolicy( QSizePolicy::Maximum, QSizePolicy::Maximum ) );
  connect( cancelBt, SIGNAL(clicked()), 
    this, SLOT(reject()));
  connect( doneBt, SIGNAL(clicked()), 
    this, SLOT(accept()));
  doneBt->setFocus();
#endif  
  
  connect( nameLE, SIGNAL(textChanged(const QString&)), 
    this, SLOT(onNameEdit()) );

  connect( fileLE, SIGNAL(textChanged(const QString&)), 
    this, SLOT(onFileEdit()) );

  if( mode != DP_NEW ) {
    nameLE->setEnabled( false );
    fileLE->setEnabled( false );
  }
  if( mode == DP_SHOW ) {
    descriptionMLE->setEnabled( false );
    maintainerLE->setEnabled( false );
    copyrightLE->setEnabled( false );
  }
  
  onNameEdit();
  
}

void DicProperties::onNameEdit()
{
  if( mode != DP_NEW || fileNameEdited ) return;
  
  QString baseName = nameLE->text().lower().simplifyWhiteSpace();

  //Replace space with underscore
  int pos = 0;
  while( true ) {
    pos = baseName.find( " ", pos );
    if( pos == -1 ) break;
    baseName = baseName.replace( pos, 1, "_" );
  }
  fileLE->setText( dicPath + "/" + baseName + QString( ".edic" ) );
}

void DicProperties::onFileEdit()
{
  if( !fileLE->hasFocus() ) return; // Only if manually edited
  fileNameEdited = !fileLE->text().isEmpty();
}


void DicProperties::onFileSelClicked()
{
  QFileInfo fi( fileLE->text() );
  
#ifdef DESKTOP
  QFileDialog fileDialog( fi.dirPath(), tr( "Bedic editable dictionaries (*.edic)" ), this );
  fileDialog.setMode( QFileDialog::AnyFile );
#else
  ZFileDialog fileDialog( tr( "Select file name for new dic" ), fi.dirPath(), ZFileDialog::AnyFile, this );
  fileDialog.setFilters( QStringList( ".edic" ) );
#endif
//   QString fileName = fileLE->text();
//   int pos = fileName.findRev( '/' );
//   if( pos != -1 )
//     fileName = fileName.right( fileName.length() - pos -1 );
  fileDialog.setSelection( fi.fileName() );

  if( fileDialog.exec() != QDialog::Accepted )
    return;
  
  QString fname = fileDialog.selectedFile();
  if( fname.right( 5 ) != ".edic" )
    fname += ".edic";
  
  fileLE->setText( fname );
//  QStringList selected = fileDialog.selectedFiles();

  fileNameEdited = true;  
}

