#ifndef HISTORY_BROWSER_H
#define HISTORY_BROWSER_H

#include <qdialog.h>
#include <qlist.h>
#include <qdatetime.h>
#include <qlistbox.h>

#include <list>

#include "abstract_config.h"


// =========== History Popup ========

struct HistoryEntry
{
  QString entry;
  QString dict;
  QDate date;
  int item_index;

  HistoryEntry( const QString &entry, const QString &dict, const QDate &date ) :
    entry( entry ), dict( dict ), date( date ), item_index( -1 )
  {
  }

  HistoryEntry()
  {
  }
  
};


typedef std::list<HistoryEntry> HistoryList;

class HistoryPopup: public QDialog
{
Q_OBJECT
  HistoryList entries;
  HistoryList::iterator selectedEntry;
  QListBox *hist_list;
  int fontHeight;
  QFont *small_font;

public:
  HistoryPopup( QWidget *parent );

  void addEntry( const QString &entry, const QString &dict );
  void updateEntries();

  void show();

  void loadHistory( AbstractConfig &config );
  void saveHistory( AbstractConfig &config );

  void centerCurrentItem();

  void setFontHeight( int height )
    {
      fontHeight = height;
    }  

  void clearHistory();
  
  const HistoryEntry getSelectedEntry()
    {
      return *selectedEntry;
    }

  void polish();
  
protected slots:
  void itemSelected( int );
  void itemSelected( QListBoxItem * item );

};


class ListBoxSeparator : public QListBoxItem
{
  QString text;
  QFont *font;
  int sep_height;
  
public:
  ListBoxSeparator( const QString &text, QFont *font, QListBox * listbox = 0 );
  int height( const QListBox * ) const;
  int width( const QListBox * ) const;
  
protected:
  void paint( QPainter * );  
};

class HistoryLBEntry: public QListBoxText
{
public:
  HistoryLBEntry( const QString &text ) :
    QListBoxText( text )
    {
    }
  
};

// =========== Browser Stack ===========

struct BrowserStackEntry
{
  QString entry;
  QString dict;
  int pos;

  BrowserStackEntry() : entry( QString::null )
  {
  }

  BrowserStackEntry( const QString &entry, const QString &dict, int pos ) :
    entry( entry ), dict( dict ), pos( pos )
  {
  }  
};

typedef std::list<BrowserStackEntry> BrowserStackList;

class BrowserStack
{
  BrowserStackList stack;
  BrowserStackList::iterator current;
  BrowserStackEntry nullEntry;
  
public:
  BrowserStack();
  
  void addEntry( const BrowserStackEntry &entry );
  const BrowserStackEntry &backEntry();
  const BrowserStackEntry &forwardEntry();
  bool hasBack();
  bool hasForward();

  void load( AbstractConfig &conf );
  void save( AbstractConfig &conf );
  
};


#endif
