/****************************************************************************
* zfiledialog.h
*
* Copyright (C) 2004 Rafal Mantiuk <rafm@users.sourceforge.org>
*
* This program is based on the kbedic dictionary by 
* Radostin Radnev <radnev@yahoo.com>.
*
* This program is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
****************************************************************************/

#ifndef ZFILEDIALOG_H
#define ZFILEDIALOG_H

#include <qdialog.h>
#include <qdir.h>
#include <qpixmap.h>

class QListView;
class QComboBox;
class QListViewItem;
class QLineEdit;

class ZFileCustomFilter
{
public:
  virtual void filter( QStringList &fileList ) = 0;
};


class ZFileDialog: public QDialog
{
Q_OBJECT
public:
  enum Mode { AnyFile, ExistingFile, Directory, ExistingFiles, DirectoryOnly };
  ZFileDialog( const QString title, const QString &path, Mode mode, QWidget *parent );
  
  const QDir* dir() const
    {
      return &currentDir;
    }

  void setFilters( const QStringList &filters );
  
  QStringList selectedFiles() const;
  QString selectedFile() const;

  void setSelection( const QString &name );

  void setCustomFilter( ZFileCustomFilter *filter )
    {
      customFilter = filter;
    }
  

  virtual void show();
  
protected:

  QDir currentDir;
  QListView *fileLV;
  QComboBox *fstabCB;

  QStringList dir_list;
  QStringList filters;

  QLineEdit *selection;

  Mode mode;

  ZFileCustomFilter *customFilter;

private:
  void insertDirEntry( const QString &label, const QString &path, QPixmap *pixmap = NULL );

protected:
  void setDir( QString path );
  void refreshFileList();
  int insertDirTree( const QString &label, QString basePath, const QString &path, bool subCheck,
    QPixmap *pixmap = NULL );
  
protected slots:
  void dirSelected( int idx );
  void fileClicked( QListViewItem *it );
  void parentDirClicked();
  void itemSelected();
  
};


#endif
